/*ATTRIB.H************************************************************* 
** ALang Execution/Mutation Engine: Version 0.6 >> %1
** Part of Project von Nuemann
** Copyright Dr I G Badcoe (I.Badcoe"bristol.ac.uk),
**           Toby Dickenson (htrd@tcp.co.uk) and
**           the Directors of Project von Nuemann.
**
** License is granted for the use of this software only as part of
** Project von Nuemann.
*/

#ifndef _ATTRIB_H_
#define _ATTRIB_H_

/*
** These bits can be combined to represent type-based or similar
** inter-node dependencies. Please keep this to 15 bits wide for
** the benefit of us testing on 16 bit compilers still :-(
*/

enum AL_Attrib_Code {
  /* TYPE : Integer or Object */
  INT_ATTR = 0x0001,
  OBJ_ATTR = 0x0002,

  /* MULTIPLICITY : Single value or list */
  SING_ATTR = 0x0004,
  LIST_ATTR = 0x0008,

  /* FLAVOUR : Return-Value, Variable, Gene or LDF */ 
  RET_ATTR = 0x0010,
  VAR_ATTR = 0x0020,
  GENE_ATTR = 0x0040,
  LDF_ATTR = 0x0080,

  /* REFERENCE : Pointer or Reality */
  PNT_ATTR = 0x0100,
  VAL_ATTR = 0x0200,
  
  /* Sub-tree is non-constant (won't always return same value). */
  NONK_ATTR = 0x0400,

  /* Sub-tree below this point contains a 'local' feature
  ** (e.g: variable or LDF) and so can't be passed to another
  ** gene.  */
  LOCAL_ATTR = 0x0800,

  /* SIGNALS : for talking to the mutagenesis routine. */
  DEL_ATTR = 0x1000,	/* could delete this child */
  END_ATTR = 0x2000,	/* could end with previous child */
  INS_ATTR = 0x4000,	/* could insert an extra child here */

  /* We can set this if we ever need to pass an explicit error 
  ** value anywhere (there are combinations of the above that
  ** are implied errors (such as 0x0000 or 0x0003) as well).  */
  ERROR_ATTR = 0x8000
};

/* To get around awkwardnesses with including parameter types in a genes
** type we'll treat an attrib as an array of ints.  */

#define MAX_PARMS 16
#define ATTRIB_SIZE (sizeof(int) * (MAX_PARMS + 2))
#define DECL_ATTRIB(a) int a[ATTRIB_SIZE]
typedef int* AL_Attrib;

/* The way this will work will be that a call to a ResFun or an AttrFun
** will pass an 18-int buffer which will be used for the return value.
** If the type is not a gene, then the last 17 ints are irrelevent
** (undefined).  If the type is a gene, then the second int is a parameter
** count and the rest are the parameter types.
**
** Note that this limits a gene to 16 parameters.
** <<Should be enough ? <IGB>>>  */

/* Expects an int so you should use []'s on the AL_Attrib (int*),
** to get the part of the type you want to modify, e.g:
**
**		LOSE_ATTR(type[3], PNT_ATTR); */

#define LOSE_ATTR(a,b) ((a) &= (~b))
#define ADD_ATTR(a,b) ((a) |= (b))

#define MASK_ATTR(a,b) ((a) &= (b))
#define IMASK_ATTR(a,b) ((a) &= (~b))

/* Compares basic type (must supply []'s),
** ignores flags and signals.
** (e.i: WON'T WORK FOR FLAGS AND SIGNALS !) */
#define ACCEPTS_ATTR(a,b) (((a) & (b) & TYPING_ATTR) == ((b) & TYPING_ATTR))
/* Compares parameter types as well. */
#define ACCEPTS_PARS(a,b) (Attrib_Compare((a), (b)))

/* 'a' and 'b' have at least one overlap */
#define HAS_SOME_ATTR(a,b) (((a) & (b)) != 0)
/* 'a' and 'b' have no overlap */
#define HAS_NONE_ATTR(a,b) (((a) & (b)) == 0)

#define COPY_ATTR(a,b) (memcpy((a), (b), ATTRIB_SIZE))

#define FLAG_ATTR (LOCAL_ATTR | NONK_ATTR)
#define NOT_FLAG_ATTR (~FLAG_ATTR)

#define SIG_ATTR (INS_ATTR | DEL_ATTR | END_ATTR | ERROR_ATTR)
#define NOT_SIG_ATTR (~SIG_ATTR)

#define TYPE_ATTR (INT_ATTR | OBJ_ATTR)
#define MULT_ATTR (SING_ATTR | LIST_ATTR)
#define FLAV_ATTR (RET_ATTR | GENE_ATTR | LDF_ATTR | VAR_ATTR)
#define REF_ATTR (PNT_ATTR | VAL_ATTR)
#define TYPING_ATTR (TYPE_ATTR | MULT_ATTR | FLAV_ATTR | REF_ATTR)
#define NON_TYPING_ATTR (~TYPING_ATTR)

#define MIX_FLAGS(to, from) ((to)[0] |= ((from)[0] & FLAG_ATTR))

/* The first 4 bits of the attribute are used to define basic data types.
** (first two: TYPE, second two: MULTIPLICITY)
**
** The next 4 (FLAVOUR) indicate different flavours for that type.
** RET_ATTR = an actual value return
** VAR_ATTR = a variable of that type
** GENE_ATTR = a gene returning that type (note that the parameter
**             types are appended to this, see bottom of file).
** LDF_ATTR = an LF returning that type
**
** The next two disctinguish between pointers and actualities.
**
** Some combinations of FLAVOUR and REFERENCE are meaningless (as yet):
** RET_ATTR  | PNT_ATTR : pointer to a return value (never happens)
** VAR_ATTR  | VAL_ATTR : value of a variable
**                        (appears as RET_ATTR | VAL_ATTR of GET-node)
** the others are interpretted as:
** RET_ATTR  | VAL_ATTR : normal return value
** VAR_ATTR  | PNT_ATTR : pointer to a variable (for use by SET/GET ...)
** GENE_ATTR | VAL_ATTR : declaration of a gene
** GENE_ATTR | PNT_ATTR : pointer to a gene (for use by GCALL ...)
** LDF_ATTR  | VAL_ATTR : declaration of an LDF
** LDF_ATTR  | PNT_ATTR : pointer to an LDF (for use by LCALL ...)
**
** <<	An important point, restrictions (RES(ret_type, where, which, buffer)
**	 and attributes (ATTR(where, buffer) handle ambiguity in different ways,
**	a restriction sets multiple bits (e.g INT_ATTR | OBJ_ATTR means "int or
**	obj").  Attributes *generally* are not ambiguous and return "complete,
**	unique types" which means:
**		exactly one bit in TYPE-ATTR,
**		exactly one bit in MULT-ATTR,
**		exactly one bit in FLAV-ATTR,
**		exactly one bit in REF-ATTR.
**	When (rarely) an attribute is ambiguous, it is achieved by *ommiting*
**	a whole part of the type (see EMPTY_attr in list_n.c).  Obviously,
**	this is more limited than for restrictions but it's OK for the only
**	case I found so far. >>
** 
** The rest are not part of the type but are additional information
** NONK_ATTR = indicates that the sub-tree is non-constant.  When unset,
** this means that the mutation function could have a chance to 'collapse'
** it into a single node if it wanted (e.g: "+(1,*(3,-(2,4)))" ==> -5).
** This could even apply to some pointers:
** if(1, i_Fish, i_Frog) ==> i_Fish (yes, you should be able to put pointers
** of the same type in conditionals (but *NO* arithmetic !)).
**
** LOCAL_ATTR, is also property of a sub-tree that doesn't affect its
** type.  This one must be passed down the tree being OR'd together
** at all nodes.  The purpose is that this should be set for a node
** if, anywhere in the lower tree, a 'local feature' was used.
** That is, a feature that could not be indiscriminately passed
** to another gene, such as one of this genes variables (which are
** all local) or one of this genes sub-genes (except sometimes you
** could do the latter so I'll think that through a bit more).
**
** NONK_ATTR and LOCAL_ATTR can be correctly propagated up a tree
** using the 'MIX_FLAGS' macro in AttFun's, e.g:
**		MIX_FLAGS(ret, CHILD(where, 0));
** has the effect of merging the flags of child#0 into the value to be
** returned.
**
** The next four are SIGNALS: DEL_ATTR, END_ATTR and INS_ATTR,
** are so the ResFun can communicate with the mutagenesis routine and
** ERROR_ATTR is for all sorts of error conditions.
**
** These should *NEVER* be set by AttrFun's.
**
** DEL_ATTR: indicates to the calling routine that the child in
** question could be deleted and the remaining children moved left
** without garbaging the node.  For example, if a node required:
** (int, obj, obj, ...) then the first child's restriction would
** not specify DEL_ATTR as the second could not replace the 'int'.
** For the rest of the nodes, DEL_ATTR would be specified as the 'obj's
** could all move over one with not trouble.  The mutation routine could
** (rarely) delete a non-deleteable node anyway as in that case the
** non-matching node would simply be forced to cross over.  If DEL_ATTR
** is returned with no type attributes then it means that the node *must*
** be deleted (this might always be covered by END_ATTR but the facility
** is there ...).
**
** END_ATTR:  If this is returned on its own, then the previous node *must*
** be the final one and any remaining ones are dropped.  If it is returned
** with other options then the previous node *could* be the final one.
** Normally the mutation engine would ignore this latter case *unless* it
** had run out of nodes in the parent but rarely it could truncate the
** children.
**
** INS_ATTR: similar to DEL_ATTR but indicates that we could insert
** an extra child here if we wished, for example, where-ever a node has
** only one type of child, we could insert a new one in the middle,
** even if this meant 'spilling' one from the end of a fixed-number of
** children.  This would not be returned in places where there is
** a requirement for specific types in specific places further down
** the list, e.g:  CASE(int, any, any, any ...)
** INS_ATTR would not be returned for the first child because
** inserting an int at that position would push the existing int onto
** the start of the any's.  Since the any's are required to be the same
** type this wouldn't be wise.  INS_ATTR would be returned for the rest
** of the childen as we can have any number of them.  When INS_ATTR us
** returned, the type attributes specified with it indicate the
** restrictions on the insertion (as with a Xover).
**
** -----------------------------------------------------------------------
**
** Note that the presence of an attribute in a restriction does *require*
** it for a child node, it *permits* it.  Phrased differently, the effect is
** to forbid it when the attribure is not present.  This is obvious when we
** consider:
**		INT_ATTR | OBJ_ATTR
** which means "int or obj" and not "int and obj" !
**
** This does mean that if you leave a type of attribute out of a restriction,
** there won't be any matches, e.g. the example above taken in full would
** mean:
**
**	not single or list (= {})
**	not constant, variable, gene or ldf (= {})
**	not pointer or value (= {})
**	and not local
**
** which leaves the empty set.
**
** To help with this I'm defining some handy macros below:  */

#define ANY_FLAV (VAR_ATTR | VAL_ATTR | LDF_ATTR | GENE_ATTR)
#define ANY_TYPE (INT_ATTR | OBJ_ATTR)
#define ANY_REF  (VAL_ATTR  | PNT_ATTR)
#define ANY_MULT (SING_ATTR | LIST_ATTR)

#define ANY_VAL (ANY_TYPE | ANY_MULT | ANY_REF | VAL_ATTR)
#define ANY_GENE (ANY_TYPE | ANY_MULT | ANY_REF | GENE_ATTR)
#define ANY_LDF (ANY_TYPE | ANY_MULT | ANY_REF | LDF_ATTR)
#define ANY_VAR (ANY_TYPE | ANY_MULT | ANY_REF | VAR_ATTR)

#define ANY_GLOBAL (ANY_TYPE | ANY_MULT | ANY_FLAV | ANY_REF)
#define ANY (ANY_GLOBAL | LOCAL_ATTR)

#define I_ATTR (INT_ATTR | SING_ATTR | VAL_ATTR | RET_ATTR | LOCAL_ATTR)
#define O_ATTR (OBJ_ATTR | SING_ATTR | VAL_ATTR | RET_ATTR | LOCAL_ATTR)

#define LI_ATTR (INT_ATTR | LIST_ATTR | VAL_ATTR | RET_ATTR | LOCAL_ATTR)
#define LO_ATTR (OBJ_ATTR | LIST_ATTR | VAL_ATTR | RET_ATTR | LOCAL_ATTR)

int Attrib_Compare(AL_Attrib a, AL_Attrib b);
char *attrib_name(int t);
int type_from_name(char *name);

#endif /* _ATTRIB_H_ */
