/*EXEC.C************************************************************* 
** ALang Execution/Mutation Engine: Version 0.6 >> %1
** Part of Project von Nuemann
** Copyright Dr I G Badcoe (I.Badcoe"bristol.ac.uk),
**           Toby Dickenson (htrd@tcp.co.uk) and
**           the Directors of Project von Nuemann.
**
** License is granted for the use of this software only as part of
** Project von Nuemann.
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "attrib.h"
#include "stack.h"
#include "list.h"
#include "exec.h"
#include "gene.h"
#include "nodetype.h"
/* #include "tab.h" */
#include "memory.h"
#include "macros.h"

#include "io.h"

struct AL_Engine *CurrentProcess;

/*
** Code for executing execution trees.
*/

/*
** Stack Structure Rules:
**
** On entry to a new node (Exec_Next):
**
**(FP is set such that PPOP will restore the
** previous FP.  It thus points one space beyond
** this.)
**        (i) FP - 2 is empty for return value                        (IRETVAL)
**        (ii) FP - 1 is value of old FP
**        (iii) FP is address of node to execute,                        (THISNODE)
**        (iv) FP + 1 is status of current node                        (STATUS)
**                (0 = entered from above,
**                 +ve depends on node details,
**                 -ve might eventually be error conditions ?)
**        (v) FP + 2 and up are the values of any args        (IARG/PARG)
**                already evaluated.
**
** On return to higher node (Exit_Node):
**        (i) return value must already have been copied onto FP - 1
**        (ii) nothing above FP - 1 must be wanted
**
**
** Stack sketch:
**
** ...
** ^
** local values of previous node
** v
** ^
** returns from older brothers/sisters (what do you mean you can't tell the
**                                      sex of a node !)
** v
** our return value
** old FP
** node                <- FP points here
** our status
** ^
** our local variables
** v
** ^
** returns from any children already evaluated
** v
**                 <- SP points here
** ...
**
** Changes for version 0.5: Stack values are stored in an array large enough
** for many stack frames, however the largest programs may occasionally
** require a very deep stack, and we do not want to have to allocate
** so much space for every program. Also, It is not possible to realloc the
** stack array since we may have pointers into it.
** To get round this problem I have split stacks up into segments. We now
** allocate a smaller stack array; when this array is full we simply overflow
** into a new array. The first stack frame in the new array is slightly
** different since it is responsible for switching back to the original
** stack array. The following diagram is a comparison of an ordinary
** stack frame and a special stack frame before the Exit_Node function is
** called:
**              Ordinary          Special
**              .                  old SP
**              .                  old FP
**              return value       return value
**              old FP             Zero           <-- SP
**              UNUSED             UNUSED
*/

/*
** Initialises an AL_Engine,
** may want to push a default ADF frame onto the stack at this point.
** For certain we set up the first Gene at this point
*/

struct AL_Engine *Init_AL_Engine(struct AL_Gene * gene) {
  struct AL_Engine *which = malloc(sizeof(struct AL_Engine));
  if (!which) {
    return NULL;
  }
  CurrentProcess = which;

  which->NestedStackSegments = 0;
  which->Stack = stack_segment_malloc();
  which->Stack->Link = NULL;

  which->StackPointer = which->Stack->Stack;
  which->VarPointer = NULL;
  which->ParPointer = NULL;
  which->FramePointer = NULL;

  which->CurrentGene = NULL;
  
  which->CurrentList.nentries = 0;
  which->CurrentList.first = NULL;
  
  Enter_Gene(gene);

  return which;
}

int Close_AL_Engine(struct AL_Engine * which) {
  int ret;
  struct AL_Stack_Segment *next_segment;

  Exit_Gene();
  ret = which->Stack->Stack[2].i;

  while (STACK_SEGMENT) {
    next_segment = STACK_LINK;
    stack_segment_free(which->Stack);
    STACK_SEGMENT = next_segment;
  }

  append_list(&FreeEntries, &which->CurrentList);
  free(which);

  return ret;
}

/*
** Sets up stack for entry to a lower node:
**
**                (i) Push 0 (for return value slot)
**                (ii) Push FP
**                (iii) Set FP = SP
**                (iv) Push pointer to new node;
**                (v) Push 0 (status of any node on entry)
*/

void Enter_Node(struct AL_Node * which) {
  union AL_Stack_Value *old_sp;
  struct AL_Stack_Segment *old_stack_segment;
  /* In most cases this test fails and we skip right past this block */
  if ((SP + AL_STACK_SWAP_OVERHEAD) >
      (STACK + AL_STACK_GRANULARITY - AL_SAFETY_LIMIT)) {
    /* Set up a new empty StackSegment */

    if (++CurrentProcess->NestedStackSegments >
        AL_ULTIMATE_STACK_DEPTH / AL_STACK_GRANULARITY) {
      /* Too many stack segments.  For now just abort */
      puts("Massive stack overflow");
      abort();
    }

    old_stack_segment = STACK_SEGMENT;
    STACK_SEGMENT = stack_segment_malloc();
    STACK_LINK = old_stack_segment;

    old_sp = SP;
    SP = STACK;

    /* AL_STACK_SWAP_OVERHEAD is the count of items pushed here */
    PPUSH(old_sp);
    PPUSH(FP);

    FP = 0;                        /* Zero is key value used in Exit_Node */
  }
  IPUSH(0);
  PPUSH(FP);
  FP = SP;
  PPUSH(which);
  IPUSH(0);
#ifdef _IGB_DEBUG_
  printf("entered node: %p (a %s), frame @ %p\n", which, NAME(which),
         FP);
  fflush(stdout);
#endif
}

/*
** Performs Tidy-up when leaving a node in the upwards direction.
** Discards any items above the FP on the stack and restores
** the FP to what it was for the previous node.
**
** This leaves the return value as the top item on the stack.
*/

void Exit_Node() {
  union AL_Stack_Value return_value;
  struct AL_Stack_Segment *old_stack_segment;
  SP = FP;
  FP = PPOP();

  /* In most cases this test is false and we skip right past. */
  if (!FP) {
    /* Occasionally we need to unwind a stack segment.
    ** If there are no more stack segments then this must be
    ** the end of the program
    */
    if (CurrentProcess->NestedStackSegments) {
      --CurrentProcess->NestedStackSegments;

      return_value = VPOP();
      FP = PPOP();
      SP = PPOP();
      VPUSH(return_value);

      old_stack_segment = STACK_SEGMENT;
      STACK_SEGMENT = STACK_LINK;
      stack_segment_free(old_stack_segment);
    } else {
      /* End of the program, do nothing more and leave FP as NULL */
    }
  }
  /* ... skip right down to here */
#ifdef _IGB_DEBUG_
  printf("exited a node frame now @ %p\n", FP);
  fflush(stdout);
#endif
}

/*
** Executes one step of the current node
*/

int Exec_Next() {
  if (FP) {
#ifdef _IGB_DEBUG_
    printf("executing node: %p (a %s), frame @ %p\n",
            FP->p,
            NAME((struct AL_Node *)(FP->p)),
            FP);
    printf("\tRET =\t (%p : %i)\n", PRETVAL, IRETVAL);
    printf("\told FP =\t (%p)\n", (FP - 1)->p);
    printf("\tSTATUS =\t (%i)\n", STATUS);
    printf("\tARG(0) =\t (%p : %i)\n", PARG(0), IARG(0));
    printf("\tARG(1) =\t (%p : %i)\n", PARG(1), IARG(1));
    printf("\tARG(2) =\t (%p : %i)\n", PARG(2), IARG(2));
    printf("\tARG(3) =\t (%p : %i)\n", PARG(3), IARG(3));
    printf("\tARG(4) =\t (%p : %i)\n", PARG(4), IARG(4));
    fflush(stdout);
#endif
    EXE((struct AL_Node *)FP->p);
    return 1;
  } else {
    return 0;
  }
}

/***************************************************************************/
/************************* Test Routine ************************************/
/***************************************************************************/

void Walk_AL_NodeType(void) {
  int i;
  printf("This version recognises the following AL_Nodes:\n");
  for (i = 0; i < NumAL_NodeTypes; i++) {
    printf("%s, ", AL_NodeTypes[i]->name);
  }
  puts("");
}
