/*GEN_N.C************************************************************* 
** ALang Execution/Mutation Engine: Version 0.6 >> %1
** Part of Project von Nuemann
** Copyright Dr I G Badcoe (I.Badcoe"bristol.ac.uk),
**           Toby Dickenson (htrd@tcp.co.uk) and
**           the Directors of Project von Nuemann.
**
** License is granted for the use of this software only as part of
** Project von Nuemann.
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "attrib.h"
#include "stack.h"
#include "list.h"
#include "exec.h"
#include "nodetype.h"
#include "tab.h"
#include "memory.h"
#include "macros.h"

#include "vars.h"
#include "io.h"
#include "mutate.h"

/***************************************************************************/
/************************** General Functions ******************************/
/***************************************************************************/

/*
** These are functions that handle the 'ordinary' AL_Nodes, and can
** be loaded into the AL_NodeType if no 'special behaviour' is required.
**
** Examples of things requiring special behaviour:
**        Other forms of AI,
**        Constants, Variables
**        (possibly top-nodes if we need some sort of 'deleting a genome'
**         behaviour)
*/

/*
** Serialise an ordinary AL_Node'
*/

void GEN_ser(struct AL_Node * where, FILE * f,
              struct AL_Tab * prev_tab) {
  int namelen;
  namelen = strlen(NAME(where)) + 1;
  fputs(NAME(where), f);

  Ser_Child_List(f, prev_tab, namelen, NUMBER(where), &CHILD(where, 0));
}

void Ser_Child_List(FILE * f, struct AL_Tab * prev_tab,
                    int namelen, EINT32 number_of_children,
                    struct AL_Node ** children) {
  struct AL_Tab tab;
  if (number_of_children == 0) {
    fputs("()", f);
  } else if (number_of_children == 1) {
    fputs("(", f);
    FTab_Unary(children[0], &tab, f, namelen, prev_tab);
    SER(children[0], f, &tab);
    fputs(")", f);
  } else {
    int i;
    fputs("(", f);
    FTab_First(children[0], &tab, f, namelen,
               prev_tab);
    for (i = 0; i < number_of_children - 1; i++) {
      if (i > 0)
        FTab_Middle(&tab, f);
      SER(children[i], f, &tab);
      fputs(",", f);
    }
    FTab_Last(&tab, f);
    SER(children[number_of_children - 1], f, &tab);
    fputs(")", f);
  }
}

/* Convert a file to an ordinary AL_Node, Deserialise_AL_Node has already read
** and identified the node-name from the file.  It then calls the nodes
** 'personal' DSER function which is either this or a 'specialised'
** function as approppriate.  */

int GEN_dser(struct AL_Node * where, FILE * file) {
#ifndef NDEBUG
  char msg[80];
#endif

  if (DSer_Child_List(file, &NUMBER(where),
                      (struct AL_Node ***)&(PRIV(where)))) {
    return 1;
  } else {

#ifndef NDEBUG
    sprintf(msg, "Error in deserializing %s", where->my_type->name);
    TRACEBACK(msg);
#endif

    return 0;
  }
}

/* Deserialize a list of nodes, usually meaning run through all the children
** of some other node.
** Removed the folloing bit:
**
**#ifndef NDEBUG
**      sprintf(msg, "Error in deserializing child %d", *total);
**      TRACEBACK(msg);
**#endif  
**      free(*children);
**      return 0;
** because NULL returns are no-longer an error condition.  We now need them
** so that we can discard gene prototypes from the tree.
**
** In order to still have error returns I've changed DSer_AL_Node to
** return an int error code and stuck the Node pointer return into the
** parameter list.  */

int DSer_Child_List(FILE * file,
                    int *total,
                    struct AL_Node *** children) {
  int c;

  *total = 0;
  *children = NULL;

  while (1) {
    c = my_fgetc(file);
    if (c == EOF) {
      TRACEBACK("Unexpected end of file");
      free(*children);
      return 0;
    }
    if (c == ')') {
      return 1;
    }
    if (c != ',' && c != '(') {
      TRACEBACK("\',\' or \'(\' missing.");
    }
    /* Clumsy but will work for present. */
    *children = realloc(*children,
                        (size_t) (sizeof(struct AL_Node *) * (1 + *total)));
    if (!DSer_AL_Node(file, (*children) + (*total))) {
      return 0;
    }
    if ((*children)[*total] != NULL) {
      (*total)++;
	}   /* if that wasn't true then we have just 'stepped over' a
        ** gene-prototype.  Which generates a CURRENT_GENE->Genes[x] entry
        ** but no tree.  */
  }
}

/* Delete an ordinary AL_Node.  */

void GEN_del(struct AL_Node * where) {
  int i;
  for (i = 0; i < NUMBER(where); i++) {
    DEL(CHILD(where, i));
  }

  free(PRIV(where));
  AL_Node_free(where);
}

int GEN_iter(struct AL_Node * where,
              CheckFun * check,
              void *private) {
  int i;
  if (!check(where, private))
    return 0;

  for (i = 0; i < NUMBER(where); ++i)
    if (!ITER(CHILD(where, i), check, private))
      return 0;

  return 1;
}

/*
** General attribute functions for simple cases.
*/

/* Two Ints */

int II_res(AL_Attrib ret_type,
           struct AL_Node * where,
           int i,
           AL_Attrib buffer) {
  if (!ACCEPTS_ATTR(ret_type[0], INT_ATTR | SING_ATTR | RET_ATTR | VAL_ATTR)) {
    return 0;
  }
  switch (i) {
    case 0:
    case 1:
      buffer[0] = INT_ATTR | SING_ATTR | RET_ATTR | VAL_ATTR |
                  DEL_ATTR | INS_ATTR;
      break;
    case 2:
      buffer[0] = END_ATTR;
      break;
    default:
      buffer[0] = ERROR_ATTR;
      break;
  }
  return 1;
}

/* Any number of ints. */

int MI_res(AL_Attrib ret_type,
           struct AL_Node * where,
           int i,
           AL_Attrib buffer) {
  if (!ACCEPTS_ATTR(ret_type[0], INT_ATTR | SING_ATTR | RET_ATTR | VAL_ATTR)) {
    return 0;
  }
  buffer[0] = INT_ATTR | SING_ATTR | RET_ATTR | VAL_ATTR | 
              INS_ATTR | DEL_ATTR | END_ATTR;
  return 1;
}

/* Two objs */

int OO_res(AL_Attrib ret_type,
           struct AL_Node * where,
           int i,
           AL_Attrib buffer) {
  if (!ACCEPTS_ATTR(ret_type[0], OBJ_ATTR | SING_ATTR | RET_ATTR | VAL_ATTR)) {
    return 0;
  }
  switch (i) {
    case 0:
    case 1:
      buffer[0] = OBJ_ATTR | SING_ATTR | RET_ATTR | VAL_ATTR |
                  DEL_ATTR | INS_ATTR;
      break;
    case 2:
      buffer[0] = END_ATTR;
      break;
    default:
      buffer[0] = ERROR_ATTR;
      break;
  }
  return 1;
}

/* Any number of objs. */

int MO_res(AL_Attrib ret_type,
           struct AL_Node * where,
           int i,
           AL_Attrib buffer) {
  if (!ACCEPTS_ATTR(ret_type[0], OBJ_ATTR | SING_ATTR | RET_ATTR | VAL_ATTR)) {
    return 0;
  }
  buffer[0] = OBJ_ATTR | SING_ATTR | RET_ATTR | VAL_ATTR | 
              INS_ATTR | DEL_ATTR | END_ATTR;
  return 1;
}

int LEAF_res(AL_Attrib ret_type,
             struct AL_Node *where,
             int i,
             AL_Attrib buffer) {
  switch (i) {
    case 0:
      buffer[0] = END_ATTR;
      break;
    default:
      buffer[0] = ERROR_ATTR;
      break;
    }
  return 1;	/* if we need to check return type then we'll need
  		** more veriosns of this function. */
}

void I_attr(struct AL_Node *where, AL_Attrib buffer) {
  buffer[0] = INT_ATTR | SING_ATTR | RET_ATTR | VAL_ATTR;
}

void LI_attr(struct AL_Node *where, AL_Attrib buffer) {
  buffer[0] = INT_ATTR | LIST_ATTR | RET_ATTR | VAL_ATTR;
}

/* This function must copy the children (or whatever private data)
** from was_where into where, which are both of the same type.
** We are allowed to use nodes under node_pool as a source for
** crossover mutations.  */

int GEN_copy(struct AL_Node * where,
             struct AL_Node * was_where,
             struct AL_Node * node_pool,
             AL_Attrib ret_type) {
  return CM_children(where, was_where, node_pool, ret_type);
}
