/*IO.C************************************************************* 
** ALang Execution/Mutation Engine: Version 0.6 >> %1
** Part of Project von Nuemann
** Copyright Dr I G Badcoe (I.Badcoe"bristol.ac.uk),
**           Toby Dickenson (htrd@tcp.co.uk) and
**           the Directors of Project von Nuemann.
**
** License is granted for the use of this software only as part of
** Project von Nuemann.
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "attrib.h"
#include "stack.h"
#include "list.h"
#include "exec.h"
#include "gene.h"
#include "nodetype.h"
/* #include "tab.h" */
#include "memory.h"
#include "macros.h"

#include "vars.h"
#include "pars.h"
#include "io.h"

/* think these will only arise here */
static struct AL_Node *Special_DSer_K(char *name);
static struct AL_Node *Special_DSer_GCall(FILE *f, char *name);
static struct AL_Node *Special_DSer_PN(char *name);
static struct AL_Node *Special_DSer_VN(char *name, int type);

/* This is used to provide a "dummy execution environment" during file IO
** *and* mutate/copy processes.  At the moment this means keeping track of
** the gene we're in (CURRENT_GENE => CopyEngine.CurrentGene) but later I
** anticipate needing to execute code-fragments (const collapsing ...) */

struct AL_Engine CopyEngine;

/* need this 'cos strchr doesn't seem to handle ascii values above 128
** (signed/unsigned problem ?)  */

int ok_char(int c) {
  char *acceptable =
  "abcdefghijklmnopqrstuvwxyz"
  "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  "01234567890(,)+-*/|&^_:";
  if (strchr(acceptable, c))
    return 1;
  if (EOF == c)
    return 1;
  return 0;
}

/* Print the (formal) parameter list for a gene.  */

void AL_Ser_Pars(struct AL_VarDetail *pars,
                 int npars,
                 FILE *f) {
  int i;

  fputs("(", f);
  if (npars > 0) {
    for(i = 0; i < npars - 1; i++) {
      fprintf(f, "%s, ", pars[i].name);
    }
    fprintf(f, "%s", pars[npars - 1].name);
  }
  fputs(")", f);
}

/* Print the sub-gene prototypes for a gene.  */

void AL_Ser_Protos(struct AL_Gene *genes,
                   int ngenes,
                   FILE *f) {
  int i;

  for(i = 0; i < ngenes; i++) {
    Ser_AL_Gene(genes + i, f, 1);
    fputs(", ", f);
  }
}

/* Print the sub-gene declarations for a gene.  */

void AL_Ser_Genes(struct AL_Gene *genes,
                  int ngenes,
                  FILE *f) {
  int i;

  for(i = 0; i < ngenes; i++) {
    Ser_AL_Gene(genes + i, f, 0);
    if (i < ngenes - 1)
      fputs(", ", f);
  }
}

/* A wrapper for the gene-serialising routine 'cos there's a little
** initialisation to be done on the first entry.  */

void Ser_AL_Genome(struct AL_Gene * which,
                 FILE * f) {
  struct AL_Engine *keep;
  
  keep = CurrentProcess;
  CurrentProcess = &CopyEngine;
  
  Ser_AL_Gene(which, f, 0);

  CurrentProcess = keep;
}

/* Serialise a whole gene.
** if 'proto' is non-zero we generate only a prototype,
** otherwise a full declaration  */

void Ser_AL_Gene(struct AL_Gene * which,
                 FILE * f,
                 int proto) {
  struct AL_Gene *keep;

  keep = CURRENT_GENE;
  CURRENT_GENE=which;

  fprintf(f, "GENE:%s ", which->Name);
  AL_Ser_Pars(which->Pars, which->NumPars, f);
  fputs("(", f);
  if (!proto) {
    AL_Ser_Protos(which->Genes, which->NumGenes, f);
    SER(which->Tree, f, NULL);
    if (which->NumGenes > 0) {
      fputs(",", f);
      AL_Ser_Genes(which->Genes, which->NumGenes, f);
    }
  }
  fputs(")\n\n", f);

  CURRENT_GENE = keep;
}

/* Wrapper function for DSer_AL_Gene because there are setups
** to do in the outermost call that don't apply to recursive
** calls.
**
** Also, this one expects to come in at the start of the file so
** it has to pass control through DSer_AL_Node (to take the 'GENE'
** word off the front) first.
**
** Won't go through DSer_AL_Node any more, 'cos that will want to discard
** gene defining nodes from the tree.  In the top one, however, there is no
** parent gene to catch them so we'll do it ourselves.  */

struct AL_Node *DSer_AL_Genome(FILE * f) {
  struct AL_Engine *save_eng;
  struct AL_Node *temp_node;
  
  if (my_fgetc(f) != 'G' ||
      my_fgetc(f) != 'E' ||
      my_fgetc(f) != 'N' ||
      my_fgetc(f) != 'E' ||
      my_fgetc(f) != ':') {
    return NULL;
  }

  ungetc(':', f);

  save_eng = CurrentProcess;
  CurrentProcess = &CopyEngine;
  CURRENT_GENE = NULL;
  
  temp_node = AL_Node_malloc();
  if (!temp_node) {
    return NULL;
  }
  
  TYPE(temp_node) = &GENE_TYPE;
  NUMBER(temp_node) = 0;
  PRIV(temp_node) = DSer_AL_Gene(f);
  
  if (!PRIV(temp_node)) {
    AL_Node_free(temp_node);
    return NULL;
  }
  
  CurrentProcess = save_eng;
  return temp_node;
}

/* Deserialise a whole gene.
** The scanner calls this having read the word 'GENE' so the next
** bits in the file are a ':', the name and parameter list.
**
** just as a reminder what they look like:
**
** struct AL_Gene {
**   char *Name;
**   AL_Attrib Type;
**   struct AL_Node *tree;
**   int NumVars;
**   struct AL_VarDetail *Vars;
**   int NumPars;
**   struct AL_VarDetail *Pars;
**   int NumGenes;
**   struct AL_Gene *Genes;
**   int NumLFs;
**   struct AL_LFDetail *LFs;
**   struct AL_Gene *parent;
** }; */

struct AL_Gene *DSer_AL_Gene(FILE * f) {
  struct AL_Gene *temp_gene = NULL;
  char name[AL_NAME_MAX + 1];
  
  if (fscanf(f, ":%20s (", name) != 1) {
    TRACEBACK("Failed to find gene-name or name too long in DSer_AL_Gene");
    return NULL;
  }
  
  if (!CURRENT_GENE) {
    puts("\nStarting to DSer outer-gene:");
    temp_gene = DSer_AL_Outer(name, f);
    puts("\nFinished DSer'ing outer-gene:");
  } else {
    temp_gene = Find_Local_AL_Gene(name, CURRENT_GENE);	/* already exist ? */

    if (!temp_gene) {
      puts("\nStarting to DSer prototype-gene:");
      temp_gene = DSer_AL_Proto(name, f);
      puts("\nFinished DSer'ing prototype-gene:");
    } else {
      puts("\nStarting to DSer full-gene-definition:");
      temp_gene = DSer_AL_Defn(temp_gene, f);
      puts("\nFinished DSer'ing full-gene-definition:");
    }
  }

#ifdef _IGB_DEBUG_
  printf("GENE:%s (p:%i v:%i sg:%i l:%i)\n",
         temp_gene->Name,
         temp_gene->NumPars,
         temp_gene->NumVars,
         temp_gene->NumGenes,
         temp_gene->NumLFs);
  dump_pars(temp_gene);
  dump_vars(temp_gene);
#endif
         
  return temp_gene;
}

struct AL_Gene *DSer_AL_Outer(char *name, FILE * f) {
  int i;
  struct AL_Gene *outer;
  int numchildren;
  struct AL_Node **temp_tree;

  outer = malloc(sizeof(struct AL_Gene));
  if (!outer)
    return NULL;
                                  
  strcpy(outer->Name, name);
  outer->Tree = NULL;
  outer->NumVars = 0;
  outer->Vars = NULL;
  outer->NumPars = 0;
  outer->Pars = NULL;
  outer->NumGenes = 0;
  outer->Genes = NULL;
  outer->NumLFs = 0;
  outer->LFs = NULL;
  
  outer->Type[0] = type_from_name(outer->Name);
  if (HAS_SOME_ATTR(ERROR_ATTR, outer->Type[0])) {
    TRACEBACK("Error type for gene in DSer_AL_Gene_i.");
    Blank_AL_Gene(outer);
    return NULL;
  }
  
  /* Now read formal parameter list. */
  outer->NumPars = DSer_Params(f, &outer->Pars);
  
  /* And copy that information into the gene's type. */
  outer->Type[1] = outer->NumPars;
  for(i = 0; i < outer->NumPars; i++) {
    outer->Type[i + 2] = outer->Pars[i].type;
  }
  
  outer->Parent = NULL;
  CURRENT_GENE = outer;

  /* Read Prototypes of Sub-Genes, Body of Gene and Definitions of
  ** Sub-Genes in that order.  Prototypes and definitions dissapear
  ** from the tree structure as they are stored elsewhere.  */
  
  if (!DSer_Child_List(f, &numchildren, &temp_tree)) {
    free(temp_tree);
    Blank_AL_Gene(outer);
    return NULL;
  }

  if (numchildren > 1) {
    outer->Tree = AL_Node_malloc();
    TYPE(outer->Tree) = &ALL_TYPE;
    NUMBER(outer->Tree) = numchildren;
    PRIV(outer->Tree) = temp_tree;
  } else {
    outer->Tree = temp_tree[0];
    free(temp_tree);
  }
  
  return outer;
}

struct AL_Gene *DSer_AL_Proto(char *name, FILE *f) {
  struct AL_Gene *temp_gene;
  int i;

  CURRENT_GENE->Genes = realloc(CURRENT_GENE->Genes,
                                sizeof(struct AL_Gene) *
                                  (CURRENT_GENE->NumGenes + 1));
  if (!CURRENT_GENE->Genes) {
    TRACEBACK("Allocation failure in DSer_AL_Gene_i.");
    return NULL;
  }

  temp_gene = CURRENT_GENE->Genes + CURRENT_GENE->NumGenes;
  CURRENT_GENE->NumGenes++;
                                  
  strcpy(temp_gene->Name, name);
  temp_gene->Tree = NULL;
  temp_gene->NumVars = 0;
  temp_gene->Vars = NULL;
  temp_gene->NumPars = 0;
  temp_gene->Pars = NULL;
  temp_gene->NumGenes = 0;
  temp_gene->Genes = NULL;
  temp_gene->NumLFs = 0;
  temp_gene->LFs = NULL;
  
  temp_gene->Type[0] = type_from_name(temp_gene->Name);
  if (HAS_SOME_ATTR(ERROR_ATTR, temp_gene->Type[0])) {
    TRACEBACK("Error type for gene in DSer_AL_Gene_i.");
    Blank_AL_Gene(temp_gene);
    return NULL;
  }
  
  /* Now read formal parameter list. */
  temp_gene->NumPars = DSer_Params(f, &temp_gene->Pars);
  
  /* And copy that information into the gene's type. */
  temp_gene->Type[1] = temp_gene->NumPars;
  for(i = 0; i < temp_gene->NumPars; i++) {
    temp_gene->Type[i + 2] = temp_gene->Pars[i].type;
  }
  
  /* Since this is a prototype, there should only be empty brackets
  ** here. */
  
  if (my_fgetc(f) != '(' || my_fgetc(f) != ')') {
    TRACEBACK("Gene Definition where prototype expected.");
    exit(1);	/* not sure how to recover here */
  }
  
  return temp_gene;
}

struct AL_Gene *DSer_AL_Defn(struct AL_Gene *proto, FILE * f) {
  int i;
  int numpar;
  struct AL_VarDetail *pars;
  int numchildren;
  struct AL_Node **temp_tree;

  if (proto->Tree) {				/* already defined ? */
    TRACEBACK("warning: gene defined more than once:");
    TRACEBACK(proto->Name);

    DEL(proto->Tree);				/* scrap existing defn's */
    proto->Tree = NULL;
    free(proto->Vars);
    proto->NumVars = 0;
    for(i = 0; i < proto->NumGenes; i++) {
      Blank_AL_Gene(proto->Genes + 1);		/* must do to remove sub-genes */
    }
    free(proto->Genes);
    proto->NumGenes = 0;
    free(proto->LFs);
    proto->NumLFs = 0;
  }
  
  proto->Parent = CURRENT_GENE;
  CURRENT_GENE = proto;

  proto->Type[0] = type_from_name(proto->Name);
  if (HAS_SOME_ATTR(ERROR_ATTR, proto->Type[0])) {
    TRACEBACK("Error type for gene in DSer_AL_Defn.");
    Blank_AL_Gene(proto);
    return NULL;
  }
  
  /* Now read formal parameter list. */
  numpar = DSer_Params(f, &pars);

  if(numpar != proto->NumPars) {
    TRACEBACK("different number of parameters between gene prototype");
    TRACEBACK("and definition");
    TRACEBACK(proto->Name);
    Blank_AL_Gene(proto);
    return NULL;
  }
  
  for(i = 0; i < proto->NumPars; i++) {
    if (proto->Pars[i].type != pars[i].type) {
      TRACEBACK("gene parameter type changed");
      Blank_AL_Gene(proto);
      return NULL;
    }
  }
  
  /* Read Prototypes of Sub-Genes, Body of Gene and Definitions of
  ** Sub-Genes in that order.  Prototypes and definitions dissapear
  ** from the tree structure as they are stored elsewhere.  */
  
  if (!DSer_Child_List(f, &numchildren, &temp_tree)) {
    free(temp_tree);
    Blank_AL_Gene(proto);
    return NULL;
  }

  if (numchildren > 1) {
    proto->Tree = AL_Node_malloc();
    TYPE(proto->Tree) = &ALL_TYPE;
    NUMBER(proto->Tree) = numchildren;
    PRIV(proto->Tree) = temp_tree;
  } else {
    proto->Tree = temp_tree[0];
    free(temp_tree);
  }
  
  return proto;
}

/* Deserialize the formal parameter list of a gene. */

int DSer_Params(FILE *f, struct AL_VarDetail **where) {
  int i, type, num_pars;
  char name[AL_NAME_MAX + 1];
  char c;
    
  *where = NULL;
  num_pars = 0;
  
  while (1) {
    i = 0;
    while(1) {
      c = my_fgetc(f);
      if (c != ',' && c != ')') {
        name[i] = c;
        i++;
        if (i >= AL_NAME_MAX) {
          TRACEBACK("name to long in parameter list:");
          TRACEBACK(name);
          free(*where);
          return 0;
        }
      } else {
        name[i] = '\0';
        break;
      }
    }
    TRACEBACK(name);

    if (name[0] == '\0') {
      if (c == ',') {
        TRACEBACK("missing name in parameter list");
        continue;
      }
      return 0;					/* no params */
    }
    type = type_from_name(name);
    if (!ACCEPTS_ATTR(INT_ATTR | OBJ_ATTR |	/* parameter *names* are all */
                      SING_ATTR | MULT_ATTR |	/* variables */
                      VAR_ATTR |
                      VAL_ATTR, type)) {
      TRACEBACK("Unallowed type for gene parameter.");
      free(*where);
      return 0;
    }
    *where = realloc(*where, (num_pars + 1) * sizeof(struct AL_VarDetail));
    /* need test/nice allocation .... */
    (*where)[num_pars].type = type;
    strcpy((*where)[num_pars].name, name);
    num_pars++;
    if (c == ')') {
      return num_pars;
    }
  }
}

int my_fgetc(FILE * f) {
  int c;
  /* Scan past whitespace */
  for (;;) {
    c = fgetc(f);
    if (ok_char(c)) {
      break;
    }
  }

  printf("%c", c);
  return c;
}

/* Slow, but easily extensible scanner. This function will be called by
** deserialization my_type to decode an AL_Node from a file, and in turn it
** recursively calls the deserialization my_type for child AL_Nodes. This
** function uses a linear search through the linked list of AL_NodeFunction
** blocks. When the set of AL_Node types is stable this can be replaced with
** a faster search if speed is a problem.
** This may return NULL (in the event of an error ? <IGB>).
**
** OK, finally getting around to 'extensing' (tm) the scanner, as follows:
**	(i) words checked against known node names,
**	(ii) then they're checked for numbers (constants)
**	(iii) then they're checked for being well-formed names
**	(iv) then, depending on the type defined by the name, either:
**		they're checked against known Genes
**		they're checked against known LF's (not yet)
**		they're checked against the Parameters of the current gene
**		they're checked against known Variables
**
**		for LF's or Variables that don't exist, we create them
**		but Genes must be prototyped and it is an error if they don't
**		exist.  */


int DSer_AL_Node(FILE * f, struct AL_Node **where) {
  static char name[AL_NAME_MAX + 1];
  static int i, c;
  struct AL_Node *node = NULL;
  int symbol_type, is_a_number, temp;

  *where = NULL;
  
  /* Scan up to any of "(),:" */
  i = 0;
  is_a_number = 1;
  for (;;) {
    if (i >= AL_NAME_MAX) {
      TRACEBACK("Name too long");
      return 0;
    }

    c = my_fgetc(f);
    if (c == EOF) {
      break;			/* Probably now an error but leave for moment. */
    } else if (c == '(' ||	/* start of child-list or parameters */
               c == ',' || 	/* if the node has no child-list (vars/K's) */
               c == ')' ||	/*	',' or ')' will be the next char */
               c == ':') {	/* separate GENE from gene-name */
      break;
    }

    name[i++] = c;
    if (!strchr("0123456789", c)) {
      is_a_number = 0;
    }
  }

  ungetc(c, f);			/* other routines will need the delimiter */

  if (c == EOF)
    return 0;

  if (i == 0) {
    *where = NULL;
    return 1;			/* empty word, for example from "()" */
  }

  name[i] = '\0';

  if (is_a_number) {
    node = Special_DSer_K(name);
    if (!node)
      return 0;
    *where = node;
    return 1;
  }

  /* Search for AL_Node types matching name */
  for (i = 0; i < NumAL_NodeTypes; i++) {
    if ((!stricmp(name, AL_NodeTypes[i]->name))) {
      /* Conversation deleted to aid readability :-) <IGB> */
      node = AL_Node_malloc();
      if (node) {
        TYPE(node) = AL_NodeTypes[i];
        if (!(temp = DSER(node, f))) {
          AL_Node_free(node);
          node = NULL;
        } else if (temp == 2) {	/* signal that dserialisation was successful */
          node = NULL;		/* but produced no child, this happens for */
        }			/* gene definitions/prototypes, for example */
      }
      *where = node;
      return 1;
    }
  }

  symbol_type = type_from_name(name);
  if (HAS_SOME_ATTR(ERROR_ATTR, symbol_type)) {
    TRACEBACK("error: error_attribute for symbol:");
    TRACEBACK(name);
    return 0;
  }

  puts(attrib_name(symbol_type));
  
  if (HAS_SOME_ATTR(symbol_type, GENE_ATTR)) {	/* is it a gene ? */
    node = Special_DSer_GCall(f, name);
    if (!node) {
      return 0;
    }
    *where = node;
    return 1;
  }

  if (HAS_SOME_ATTR(symbol_type, VAR_ATTR)) {	/* is it a variable ? */
    node = Special_DSer_PN(name);
    if (!node) {
      node = Special_DSer_VN(name, symbol_type);
      if (!node) {
        TRACEBACK("No such variable/parameter:");
        TRACEBACK(name);
      }
    }
    *where = node;
    return 1;
  }


  /* No match */
  TRACEBACK(name);
  TRACEBACK("No match");
  return 0;
}

/* make a constant node from the a string representing the number.  */

struct AL_Node *Special_DSer_K(char *name) {
  struct AL_Node *temp;
  
  temp = AL_Node_malloc();
  if (!temp)
    return NULL;

  TYPE(temp) = &K_TYPE;
  NUMBER(temp) = atoi(name);
  PRIV(temp) = NULL;

  return temp;
}

/* make a parameter-name node from just the name,
** return NULL if there is no such parameter */

struct AL_Node *Special_DSer_PN(char *name) {
  int j;
  struct AL_Node *temp;
  
  j = get_par(name, CURRENT_GENE);
  if (j == -1) {
    return NULL;
  }
  
  temp = AL_Node_malloc();
  if (!temp)
    return NULL;

  TYPE(temp) = &PN_TYPE;
  NUMBER(temp) = j;
  PRIV(temp) = NULL;

  return temp;
}

/* make a variable-name node from just the name,
** create the variable if it does not already exist.
**
** Might ultimately want to be able to reference variables/params of
** higher genes but that will require a chain of VarPointers/ParPointers on
** the stack*/

struct AL_Node *Special_DSer_VN(char *name, int type) {
  int j;
  struct AL_Node *temp;
  
  j = get_var(name, CURRENT_GENE);
  if (j == -1) {
    j = create_var(name, CURRENT_GENE, type);
  }
  
  temp = AL_Node_malloc();
  if (!temp)
    return NULL;

  TYPE(temp) = &VN_TYPE;
  NUMBER(temp) = j;
  PRIV(temp) = NULL;

  return temp;
}

struct AL_Node *Special_DSer_GCall(FILE *f, char *name) {
  int i;
  struct AL_Gene *gene;
  struct AL_Node *node;
  struct AL_Node **temp_list;

  if ((gene = Find_AL_Gene(name))) {		/* does it exist ? */
    node = AL_Node_malloc();			/* make a gene-call node */
    if (!node) {
      return NULL;
    }
    TYPE(node) = &GCALL_TYPE;			/* ... */
    if (!DSer_Child_List(f,			/* get the parameter trees */
                         &NUMBER(node),
                         &temp_list)) {
      return NULL;
    }
    PRIV(node) = malloc((NUMBER(node) + 1) *	/* make room for param trees */
                        sizeof(struct AL_Node *));	/* *plus* name node */
    if (!PRIV(node)) {
      AL_Node_free(node);
      free(temp_list);
      return NULL;
    }
    CHILD(node, 0) = AL_Node_malloc();		/* make a gene-name-node */
    if (!CHILD(node, 0)) {
      AL_Node_free(node);
      free(temp_list);
      return NULL;
    }
    TYPE(CHILD(node, 0)) = &GN_TYPE;		/* ... */
    PRIV(CHILD(node, 0)) = gene;		/* point it at the gene */
    NUMBER(CHILD(node, 0)) = 0;

    for(i = 0; i < NUMBER(node); i++) {
      CHILD(node, i + 1) = temp_list[i];	/* copy over rest of children */
    }
    
    free(temp_list);
    NUMBER(node)++;				/* real number of children is */
						/* one more */
    return node;
  }
  
  TRACEBACK("error: unprototyped gene:");
  TRACEBACK(name);
  return NULL;
}

/* need custom memory system here and above.
** put one in memory.c eventually  */

void Free_AL_Gene(struct AL_Gene *which) {
  free(which);
}
