/*MEMORY.C************************************************************* 
** ALang Execution/Mutation Engine: Version 0.6 >> %1
** Part of Project von Nuemann
** Copyright Dr I G Badcoe (I.Badcoe"bristol.ac.uk),
**           Toby Dickenson (htrd@tcp.co.uk) and
**           the Directors of Project von Nuemann.
**
** License is granted for the use of this software only as part of
** Project von Nuemann.
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "attrib.h"
#include "stack.h"
#include "list.h"
#include "exec.h"
#include "nodetype.h"
#include "tab.h"
#include "memory.h"
#include "macros.h"

/***********************************************************************
** The purpose of this module is to implement wrappers around malloc/free
** for some structs that we will be allocating and freeing frequently.
**
** The structs treated in this way so far are AL_Nodes and AL_Stack_Segments
**
** (now added the parts of lists as well)
**
** This module gets its performance gain from not always returning
** unused memory to the pool maintained by malloc/free. Internally we
** maintain a pool of preallocated AL_Nodes and AL_Stack_Segments
**
** This is efficient because:
**
** 1) During a game there is roughly a constant number of these
**    structs in play. The pre-allocated pool is usually small
**    compared to the number of structs actually being used.
** 2) During a game there is a high turnover of these structs;
**    the preallocated pool is faster than the raw malloc/free.
**
*/



/**********************************************************************
** Keep a list of malloced blocks of stack values. When another
** block is requested, first try to supply the block from this list.
** We only malloc a new block if this list is empty.
*/

static struct AL_Stack_Segment *free_stack_segments = NULL;

struct AL_Stack_Segment *stack_segment_malloc(void) {
  struct AL_Stack_Segment *to_return;
  if (free_stack_segments) {
    /* Unhook the head of the list */
    to_return = free_stack_segments;
    free_stack_segments = to_return->Link;
  } else {
    /* Bad news; actually allocate a fresh one */
    to_return = malloc(sizeof(struct AL_Stack_Segment));
  }

  return to_return;
}

void stack_segment_free(struct AL_Stack_Segment * to_free) {
  /* Never deallocate the space; just link it to the head of the list */
  to_free->Link = free_stack_segments;
  free_stack_segments = to_free;
}

/**********************************************************************
** The same for AL_Nodes
*/

static struct AL_Node *free_nodes = NULL;

struct AL_Node *AL_Node_malloc(void) {
  struct AL_Node *to_return;
  if (free_nodes) {
    /* Unhook the head of the list */
    to_return = free_nodes;
    free_nodes = PRIV(to_return);
  } else {
    /* Bad news; actually allocate a fresh one */
    to_return = malloc(sizeof(struct AL_Node));
  }

  return to_return;
}

void AL_Node_free(struct AL_Node * to_free) {
  /* Never deallocate the space; just link it to the head of the list */
  PRIV(to_free) = free_nodes;
  free_nodes = to_free;
}


/***********************************************************************/

void AL_flush_preallocated_pool() {
  void *next;
  
  while (free_stack_segments) {
    next = free_stack_segments->Link;
    free(free_stack_segments);
    free_stack_segments = next;
  }

  while (free_nodes) {
    next = PRIV(free_nodes);
    free(free_nodes);
    free_nodes = next;
  }
  
  while(FreeEntries.first) {
    next = FreeEntries.first->next;
    free(FreeEntries.first);
    FreeEntries.first = next;
  }
  FreeEntries.nentries = 0;

  while(FreeHeads) {
    next = FreeHeads->detail.next_free;
    free(FreeHeads);
    FreeHeads = next;
  }
}

/************************************************************************/

struct AL_ListHead FreeEntries;
struct AL_ListHead *FreeHeads = NULL;

/*
** get an empty list header from the free list
** (allocate more if required)
*/

struct AL_ListHead *get_header() {
  struct AL_ListHead *temp;
  if (!FreeHeads) {
    FreeHeads = alloc_HeadBlock();
  }
  temp = FreeHeads;
  FreeHeads = FreeHeads->detail.next_free;

  temp->nentries = 0;                /* blank out the header */
  temp->first = NULL;
  temp->detail.last = NULL;

  return temp;
}

/*
** return a list header to the free list
** (leaves it full of garbage)
*/

void free_header(struct AL_ListHead * which) {
  which->detail.next_free = FreeHeads;
  FreeHeads = which;
}

/*
** get a list entry from the free list
*/

struct AL_ListEntry *get_entry() {
  struct AL_ListEntry *temp;
  if (FreeEntries.nentries == 0) {
    alloc_EntryBlock();
  }
  temp = pop_entry(&FreeEntries);
  temp->value = 0;
  return temp;
}

/*
** return a whole list (entries and header) to the free lists
*/

void free_list(struct AL_ListHead * which) {
  append_list(&FreeEntries, which);
  free_header(which);
}

/*
** allocate a new block of headers and wire up as a free list
*/

struct AL_ListHead *alloc_HeadBlock() {
  struct AL_ListHead *temp, *where;
  int i;
  where = NULL;
  for (i = AL_LIST_HEAD_BLOCK - 1; i >= 0; i--) {
    temp = (struct AL_ListHead *) malloc(sizeof(struct AL_ListHead));
    if (!temp) {
      TRACEBACK("allocation failure in alloc_HeadBlock");
      exit(1);
    }
    temp->detail.next_free = where;
    where = temp;
  }

  return temp;
}

/*
** allocate a block of entries and stuff them into FreeEntries
** (which must be empty or you use the old ones)
*/

void alloc_EntryBlock() {
  struct AL_ListEntry *temp, *where;
  int i;
  where = NULL;
  for (i = 0; i < AL_LIST_ENTRY_BLOCK; i++) {
    temp = (struct AL_ListEntry *) malloc(sizeof(struct AL_ListEntry));
    if (!temp) {
      TRACEBACK("allocation failure in alloc_EntryBlock");
      exit(1);
    }
    if (i == 0) {
      FreeEntries.detail.last = temp;
    }
    temp->next = where;
    where = temp;
  }

  FreeEntries.nentries = AL_LIST_ENTRY_BLOCK;
  FreeEntries.first = temp;
}
