/*MUTATE.C************************************************************* 
** ALang Execution/Mutation Engine: Version 0.6 >> %1
** Part of Project von Nuemann
** Copyright Dr I G Badcoe (I.Badcoe"bristol.ac.uk),
**           Toby Dickenson (htrd@tcp.co.uk) and
**           the Directors of Project von Nuemann.
**
** License is granted for the use of this software only as part of
** Project von Nuemann.
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "attrib.h"
#include "stack.h"
#include "list.h"
#include "exec.h"
#include "gene.h"
#include "nodetype.h"
#include "memory.h"
#include "macros.h"

#include "io.h"

struct AL_Node **extend_list(struct AL_Node **list, int size);


struct AL_Node *CM_AL_Node(struct AL_Node * was_where,
                                    struct AL_Node * node_pool,
                                    AL_Attrib ret_type);

static int Check_Compatible_Node(struct AL_Node * where, void *private);

static struct AL_Node *Find_Compatible_Node(AL_Attrib restrictions,
                                            struct AL_Node * node_pool);

struct AL_Node Parents;
struct AL_Gene *IOCurGene;

/* Load some default nodes so that there's always one of the required type */

void load_defaults() {
  FILE *f;

  if ((f = fopen("def_tree.gpt", "r")) == NULL) {
    puts("Cannot open default GPTree-file : def_tree.gpt");
    exit(1);
  }
  if (!DSer_AL_Node(f, &CHILD(&Parents, 0)) ||
      CHILD(&Parents, 0) == NULL) {
    puts("Problem with default GPTree-file : def_tree.gpt");
    exit(1);
  }
}

/*
** This is really the top level function of this module
*/

struct AL_Node *CM_AL_Genome(struct AL_Gene * dad,
                                      struct AL_Gene * mum) {
  struct AL_Node *temp;
  struct AL_Engine *save_engine;
  
/* Here we assemble an abstract node, a node of type PARENTS that
** contains these two genes as child nodes. And some previously
** acquired default nodes intended to provide an example of every
** type of Attribute that the program might want.  This is passed
** to Mutate_AL_Node as a source for crossover mutations.
** A PARENTS node does not support any attribs, so it will never
** be used as a crossover source itself.  */
  CHILD(&Parents, 1) = dad->Tree;
  CHILD(&Parents, 2) = mum->Tree;

/* CopyEngine is used in all non-executing procedures (load/save/mutate)
** that need to keep a track of what gene we're currently in. */
  save_engine = CurrentProcess;
  CurrentProcess = &CopyEngine;
  CopyEngine.CurrentGene = NULL;

/* Produce a (possibly) mutated copy of dad, using crossover from
** both mum and dad. This leaves mum and dad unchanged.  */
  temp = CM_AL_Node(dad->Tree, &Parents, dad->Type);

  CurrentProcess = save_engine;

  if (!PRIV(temp)) {
    free(temp);
    return NULL;
  }
  return temp;
}

/* This function _COPIES_ the node was_where (and all its children)
** or replaces it with another, compatible node.
** node_pool is a pointer to a tree from which we take samples during
** crossover mutation. At the moment this value gets passed around
** as a parameter. Please dont make it a global variable or similar
** since in the future we may have some mutation operations that
** cause node_pool to be set to a different tree.
** This function may return NULL.  (but shouldn't ? <IGB>)
**
** <IGB> copying now encompases point mutations, deletions and
** insertions but these should have already be handled before reaching here.
** This routine just supplies a single child-node as requested.  */

struct AL_Node *CM_AL_Node(struct AL_Node * was_where,
                           struct AL_Node * node_pool,
                           AL_Attrib ret_type) {
  DECL_ATTRIB(child_ret);
  struct AL_Node *where;

  where = AL_Node_malloc();
  if (!where)
    return NULL;

  ATTR(was_where, child_ret);		/* attrib supplied by original. */
  if (!was_where || 				/* nothing to copy, so improvise. */
      !AL_RANDOM(4) ||				/* voluntary X-over */ 
      !ACCEPTS_PARS(ret_type, child_ret)) {	/* obligatory due to type mis-match */
						/* Crossover mutation into this child */
    was_where = Find_Compatible_Node(ret_type, node_pool);
  }

/* Now invoke the copy function specific to this node.
** This function is responsible for copying all the lower tree,
** usually by recursive calls back to this one.  */

  where->my_type = was_where->my_type;
  if (!COPY(where, was_where, node_pool, ret_type)) {
    AL_Node_free(where);
    return NULL;
  }

  return where;
}

/* This struct us used by the following two for storage of the
** nodes found.  */

struct private_for_check_compat_node {
  AL_Attrib restrictions;
  int counter;
  struct AL_Node *choice;
};

struct AL_Node **extend_list(struct AL_Node **list, int size) {
  return realloc(list, sizeof(struct AL_Node *) * size);
}

void PARENT_attr(struct AL_Node *where, AL_Attrib buffer) {
  buffer[0] = ERROR_ATTR;
  /* meaning that this node isn't suitable to go *anywhere* ! */
}

/* Copy/Mutate Children (in direct contravention of the European Declaration
** of Human Rights).  This is the 'heart' of the mutation of child-lists.
** It handles the changes to the lists and CM_AL_Node handles the X-over of
** specific members.  CM_AL_Node can be forced into an 'obligate X-over'
** if the required return type has been changed by the actions of this
** routine (or those futher up the recursion).
**
** A standard function for copying a whole set of children
** allowing for insertions/deletions/extensions and point mutations
** as permitted:
**  INSERTION = a(1,2,3) ==> a(1,4,2,3)
**  DELETION =  a(1,2,3) ==> a(1,3)
**  EXTENSION = a(1,2,3) ==> a(1,2,3,4)
**  POINT =     a(1,2,3) ==> b(1,2,3) <might also be forced to change
**                                     children> */

int CM_children(struct AL_Node *to,
                struct AL_Node *from,
                struct AL_Node *node_pool,
                AL_Attrib ret_type) {
  int i, j, more;
  DECL_ATTRIB(buffer1);
  struct AL_Node *was_where;

  i = 0;
  j = 0;
  more = 1;
  while (more) {
    RES(ret_type, from, i, buffer1);			/* "from"'s for child i.  */
    if (ACCEPTS_ATTR(buffer1[0], ERROR_ATTR)) {
      puts("Error attribute in GEN_copy_children.");
      return 0;
    }
    if (buffer1[0] == END_ATTR ||			/* must end. */
        (ACCEPTS_ATTR(buffer1[0], END_ATTR) &&		/* may end. */
         !AL_RANDOM(10))) {
      more = 0;
    } else if (buffer1[0] == DEL_ATTR ||		/* must delete. */
               (ACCEPTS_ATTR(buffer1[0], DEL_ATTR) &&	/* may delete. */
                !AL_RANDOM(10))) {
      i++;
    } else {
      PRIV(from) = extend_list(PRIV(from), j + 1);
      if (buffer1[0] != INS_ATTR ||			/* must insert. */
          (ACCEPTS_ATTR(buffer1[0], INS_ATTR) &&	/* may insert. */
           !AL_RANDOM(10)) ||
          j >= NUMBER(from)) {				/* no more originals to copy. */
        was_where = NULL;
      } else {						/* try to copy original. */
        was_where = CHILD(from, i);
        i++;
      }
      CHILD(to, j) = CM_AL_Node(was_where, 		/* X-overs at random or */
                                node_pool,		/* if changed return */
                                buffer1);		/* type requires it. */
      if (!CHILD(to, j))
        goto tidy;
      j++;
    }
  }

  return 1;

tidy:
  while (--j >= 0) {
    DEL(CHILD(to, j));
  }
  free(PRIV(to));
  return 0;
}

/* This function is used during crossover mutation. It must find
** all nodes within node_pool that comply with the restrictions and
** select one at random  */

static struct AL_Node *Find_Compatible_Node(AL_Attrib restrictions,
                                            struct AL_Node * node_pool) {
  /* This struct is a private packet of data passed down to each
  ** invokation of check_compatible_node, below  */
  struct private_for_check_compat_node s =
  {restrictions, 0, NULL};
/*  s.counter = 0;
**  s.choice = NULL;
**  s.restrictions = restrictions;  */

  ITER(node_pool, Check_Compatible_Node, &s);

  if (NULL == s.choice) {
    /* This is a 'should never happen' condition in the final game,
    ** but while developing the mutation code it actually does happen
    ** quite often. This next line is a good place to set a breakpoint;
    ** if it trips then it means that the node_pool didn't contain any
    ** nodes matching 'restrinctions'. */

    printf("Failed to find a compatible node\n");
#if 1
    return NULL;
#else
    abort();
#endif
  }
  return s.choice;
}

/* Called by ITER above */

static int Check_Compatible_Node(struct AL_Node * where, void *private) {
  struct private_for_check_compat_node *s = private;
  AL_Attrib attrib = NULL;

  ATTR(where, attrib);
  if (ACCEPTS_PARS(attrib, s->restrictions)) {
    ++s->counter;
    if (!AL_RANDOM(s->counter))
      s->choice = where;
  }
  return 1;
}
