/*GENE.C************************************************************* 
** ALang Execution/Mutation Engine: Version 0.6 >> %1
** Part of Project von Nuemann
** Copyright Dr I G Badcoe (I.Badcoe"bristol.ac.uk),
**           Toby Dickenson (htrd@tcp.co.uk) and
**           the Directors of Project von Nuemann.
**
** License is granted for the use of this software only as part of
** Project von Nuemann.
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "attrib.h"
#include "stack.h"
#include "list.h"
#include "exec.h"
#include "vars.h"
#include "gene.h"
#include "nodetype.h"
/* #include "tab.h" */
#include "memory.h"
#include "macros.h"

/*
** Gene-Entry
** 
** 	a - push the Par-pointer and set it equal to the top of the stack
** 	b - evaluate any arguments and push their return values
** 		(thus: argument-n is found at Frame-pointer + n)
**  c   push CurrentGene (old gene)
**  d - set CurrentGene = new gene
** 	e - push the Variable pointer and set it equal to the top of the stack
** 	f - push variable initial values onto the stack.  This means zero's 
**		for INT/OBJ and empty lists for the list-variables.
** 		(thus: variable-n is found at VAR-pointer + n)
** 	g - do an Enter_Node on the top node of the ADF body
**
** c, d, e, f and g are done here.
** a and b must have already been done by the call node
** as it has the argument values
*/
 
void Enter_Gene(struct AL_Gene * which) {
  int i;

  PPUSH(CURRENT_GENE);
  CURRENT_GENE = which;

  PPUSH(VP);
  VP = SP;
  
  for (i = 0; i < NVARS(which); i++) {
    init_var(VARD(which, i));
  }
  
  Enter_Node(which->Tree);
}

/* 	Exit:
** 
** 	a - copy the ADF's return value to the call nodes return value
** 	b - pop the variables, freeing lists as we go
** 	c - pop the variable pointer
**  d - pop the gene-pointer
** 	e - pop the arguments, not freeing lists 'cos they belong to the caller
** 	f - pop the PAR-pointer
** 	g - Exit_Node from the call node.
**
**  a, e, f and g must be done in the call-node.
*/

void Exit_Gene(void) {
  int i;

  for (i = NVARS(CURRENT_GENE) - 1; i >= 0; i--) {
    tidy_var(VARD(CURRENT_GENE, i));
  }

  VP = PPOP();
  CURRENT_GENE = PPOP();
}
  
/* Copy a gene, some parts are copied here, others are constructed during the
** copying of the tree:
** 
** struct AL_Gene {
**   char *Name;			<- here
**   DECL_ATTRIB(Type);			<- here
**   struct AL_Node **tree;		<- tree (obviously !)
**   int NumVars;			<- tree
**   struct AL_VarDetail *Vars;		<- tree
**   int NumPars;			<- here
**   struct AL_VarDetail *Pars;		<- here
**   int NumGenes;			<- here
**   struct AL_Gene *Genes;		<- prototypes here, definitions in tree
**   int NumLFs;			<- tree
**   struct AL_LFDetail *LFs;		<- tree
**   struct AL_Gene *Parent;		<- here
** };  */

struct AL_Gene *Copy_AL_Gene(struct AL_Gene *was_where,
                             struct AL_Node *node_pool,
                             AL_Attrib ret_type) {
  struct AL_Gene *temp_gene;
  struct AL_Gene *save_gene;
  int i;
  
  temp_gene = malloc(sizeof(struct AL_Gene));
  if (!temp_gene) {
    return NULL;
  }

  strcpy(temp_gene->Name, was_where->Name);
  
  temp_gene->NumVars = 0;
  temp_gene->Vars = NULL;

  temp_gene->NumPars = was_where->NumPars;
  COPY_ATTR(temp_gene->Type, was_where->Type);
  
  temp_gene->NumGenes = was_where->NumGenes;
  temp_gene->Genes = malloc(sizeof(struct AL_Gene) * was_where->NumGenes);
  if (!temp_gene->Genes) {
    free(temp_gene->Name);
    free(temp_gene);
    return NULL;
  }
  memcpy(temp_gene->Genes, was_where->Genes,
         sizeof(struct AL_Gene) * was_where->NumGenes);
  for(i = 0; i < was_where->NumGenes; i++) {
    temp_gene->Genes[i].Tree = NULL;
  }
  
  temp_gene->Parent = CURRENT_GENE;
  save_gene = CURRENT_GENE;

  temp_gene->Tree = malloc(sizeof(struct AL_Node *) *
                           (temp_gene->NumGenes + 1));
  if (!temp_gene->Tree) {
    free(temp_gene->Name);
    free(temp_gene->Genes);
    free(temp_gene);
    return NULL;
  }

  for(i = 0; i <= was_where->NumGenes; i++) {
    if (!COPY(temp_gene->Tree,
              was_where->Tree,
              node_pool,
              ret_type))
      goto tidy;
  }

  CURRENT_GENE = save_gene;
  return temp_gene;

tidy:
  DEL(temp_gene->Tree);
  free(temp_gene->Name);
  free(temp_gene->Genes);
  free(temp_gene);
  return NULL;
}

struct AL_Gene *Find_AL_Gene(char *name) {
  struct AL_Gene *in_which;
  struct AL_Gene *temp;

  for(in_which = CURRENT_GENE;
      in_which != NULL;
      in_which = in_which->Parent) {
    if ((temp = Find_Local_AL_Gene(name, in_which))) {
      return temp;
    }
  }
  
  return NULL;
}

struct AL_Gene *Find_Local_AL_Gene(char *name,
                                   struct AL_Gene *in_which) {
  int i;
  
  for(i = 0; i < in_which->NumGenes; i++) {
    if (!strcmp(name, in_which->Genes[i].Name)) {
      return in_which->Genes + i;
    }
  }
  
  return NULL;
}

void Blank_AL_Gene(struct AL_Gene *which) {
  int i;

  which->Name[0] = '\0';
  free(which->Pars);
  which->NumPars = 0;
  free(which->Vars);
  which->NumVars = 0;
  free(which->LFs);
  which->NumLFs = 0;
  if (which->Tree)
    DEL(which->Tree);
  for(i = 0; i < which->NumGenes; i++) {
    Blank_AL_Gene(which->Genes + i);
  }
  free(which->Genes);
  which->NumGenes = 0;
  which->Parent = NULL;
  which->Type[0] = ERROR_ATTR;
  which->Type[1] = 0;
}
